/*
 * ---------------- www.spacesimulator.net --------------
 *   ---- Space simulators and 3d engine tutorials ----
 *
 * Original Author: Damiano Vitulli
 * Porting to OpenGL3.3: Movania Muhammad Mobeen
 * Shaders Functions: Movania Muhammad Mobeen
 *
 * This program is released under the BSD licence
 * By using this program you agree to licence terms on spacesimulator.net copyright page
 *
 */

#include <gl/glew.h>
#include <gl/freeglut.h>
#include <iostream>
#include "GLSLShader.h"
#include "load_bmp.h"
#include "load_3ds.h"
#include "object.h"

//#define USE_GLM

#ifdef USE_GLM
#include <glm/glm.hpp>
#include <glm/gtc/matrix_projection.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#endif

#include <cassert>


#define GL_CHECK_ERRORS assert(glGetError()== GL_NO_ERROR);

using namespace std; 

int screen_width=640;
int screen_height=480;
const int TOTAL_OBJECTS=3;

GLuint	vboVerticesID[TOTAL_OBJECTS], 
		vboTexCoordID[TOTAL_OBJECTS], 
		 vboNormalsID[TOTAL_OBJECTS],
		 vboIndicesID[TOTAL_OBJECTS], 
			    vaoID[TOTAL_OBJECTS];

GLsizei stride = sizeof(GLfloat)*3;
GLSLShader shader;

int filling=1;

// Absolute rotation values (0-359 degrees) and rotiation increments for each frame
float rotation_x=0, rotation_x_increment=0.1f;
float rotation_y=0, rotation_y_increment=0.05f;
float rotation_z=0, rotation_z_increment=0.03f;
 
#ifdef USE_GLM
glm::mat4 
#else 	
matrix_4x4_type 
#endif
P;//projection matrix;

//Lights settings
GLfloat light_ambient[]= { 0.1f, 0.1f, 0.1f, 0.1f };
GLfloat light_diffuse[]= { 1.0f, 1.0f, 1.0f, 0.0f };
GLfloat light_specular[]= { 1.0f, 1.0f, 1.0f, 0.0f };
GLfloat light_position[]= { 100.0f, 0.0f, -10.0f, 1.0f };

//Materials settings
GLfloat mat_ambient[]= { 0.1f, 0.1f, 0.1f, 0.0f };
GLfloat mat_diffuse[]= { 1.0f, 1.0f, 1.0f, 0.0f };
GLfloat mat_specular[]= { 0.2f, 0.2f, 0.2f, 0.0f };
GLfloat mat_shininess[]= { 1.0f };

void InitShaders(void)
{
	shader.LoadFromFile(GL_VERTEX_SHADER, "shader.vert");
	shader.LoadFromFile(GL_FRAGMENT_SHADER, "shader.frag");
	shader.CreateAndLinkProgram();
	shader.Use();	
		shader.AddAttribute("vVertex");
		shader.AddAttribute("vUV");
		shader.AddAttribute("vNormal");
		shader.AddUniform("N");
		shader.AddUniform("MV");
		shader.AddUniform("MVP");
		shader.AddUniform("lightPos");
		shader.AddUniform("mat_ambient");
		shader.AddUniform("mat_diffuse");
		shader.AddUniform("mat_specular");
		shader.AddUniform("mat_shininess");
		shader.AddUniform("light_ambient");
		shader.AddUniform("light_diffuse");
		shader.AddUniform("light_specular");
		shader.AddUniform("textureMap");
		shader.AddUniform("has_texture");
		glUniform1i(shader("textureMap"),0);
		glUniform4fv(shader("lightPos"),1, light_position);

		glUniform4fv(shader("light_ambient"),1, light_ambient);
		glUniform4fv(shader("light_diffuse"),1, light_diffuse);
		glUniform4fv(shader("light_specular"),1, light_specular);

		glUniform4fv(shader("mat_ambient"),1, mat_ambient);
		glUniform4fv(shader("mat_diffuse"),1, mat_diffuse);
		glUniform4fv(shader("mat_specular"),1, mat_specular);
		glUniform1fv(shader("mat_shininess"),1, mat_shininess);
	shader.UnUse();

	GL_CHECK_ERRORS
}
void InitVAO() {
	GL_CHECK_ERRORS
	//Create vao and vbo stuff
	glGenVertexArrays(TOTAL_OBJECTS, vaoID);
	glGenBuffers (TOTAL_OBJECTS, vboVerticesID);
	glGenBuffers (TOTAL_OBJECTS, vboTexCoordID);
	glGenBuffers (TOTAL_OBJECTS, vboNormalsID);
	glGenBuffers (TOTAL_OBJECTS, vboIndicesID);

	GL_CHECK_ERRORS
	for(int i=0;i<TOTAL_OBJECTS;i++){ 
		glBindVertexArray(vaoID[i]);	 
			glBindBuffer (GL_ARRAY_BUFFER, vboVerticesID[i]);
			glBufferData (GL_ARRAY_BUFFER, sizeof(GLfloat)*3*object[i].vertices_qty, &object[i].vertex[0], GL_STATIC_DRAW);
			GL_CHECK_ERRORS
			glEnableVertexAttribArray(shader["vVertex"]);
			glVertexAttribPointer (shader["vVertex"], 3, GL_FLOAT, GL_FALSE,stride,0);
			GL_CHECK_ERRORS
			glBindBuffer (GL_ARRAY_BUFFER, vboTexCoordID[i]);
			glBufferData (GL_ARRAY_BUFFER, sizeof(GLfloat)*2*object[i].vertices_qty, &object[i].mapcoord[0], GL_STATIC_DRAW);
			GL_CHECK_ERRORS
			glEnableVertexAttribArray(shader["vUV"]);
			glVertexAttribPointer (shader["vUV"], 2, GL_FLOAT, GL_FALSE,sizeof(GLfloat)*2,0);
			GL_CHECK_ERRORS		
			glBindBuffer (GL_ARRAY_BUFFER, vboNormalsID[i]);
			glBufferData (GL_ARRAY_BUFFER, sizeof(GLfloat)*3*object[i].vertices_qty, &object[i].normal[0], GL_STATIC_DRAW);
			GL_CHECK_ERRORS
			glEnableVertexAttribArray(shader["vNormal"]);		
			glVertexAttribPointer (shader["vNormal"], 3, GL_FLOAT, GL_FALSE,stride,0);
			GL_CHECK_ERRORS
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vboIndicesID[i]);
			glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort)*3*object[i].polygons_qty, &object[i].polygon[0], GL_STATIC_DRAW);
		glBindVertexArray(0);
	}

	GL_CHECK_ERRORS
}
void InitGL() {
	MatrGenerateLookupTab();

	glGetError();
	GL_CHECK_ERRORS
	glClearColor(0.0f,0.0f,0.2f,0.0f);	
	GL_CHECK_ERRORS
	
	//Objects loading
	ObjLoad ("fighter1.3ds","skull.bmp",             -10.0, 0.0, -30.0,    900,0,0);
	ObjLoad ("fighter2.3ds",'\0',                     10.0, 0.0, -30.0,    900,0,0);
	ObjLoad ("fighter3.3ds","spaceshiptexture.bmp",    0.0, 0.0, -30.0,    900,0,0);

	GL_CHECK_ERRORS

	InitShaders();	
	InitVAO();
	
	glEnable(GL_DEPTH_TEST); // We enable the depth test (also called z buffer)
	glEnable(GL_CULL_FACE); // Enable the back face culling
	
	//glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // Texture mapping perspective correction (OpenGL... thank you so much!)
	//now shaders attributes are by default perspective correct
	GL_CHECK_ERRORS
    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL); // Polygon rasterization mode (polygon filled)	
	GL_CHECK_ERRORS    
	
	GL_CHECK_ERRORS
}

 
void OnRender() {
	GL_CHECK_ERRORS
	glClear(GL_COLOR_BUFFER_BIT| GL_DEPTH_BUFFER_BIT);
	//setup matrices
	
	for(int i=0;i<TOTAL_OBJECTS;i++) {
		#ifdef USE_GLM
		glm::mat4 MV = glm::mat4(object[i].matrix[0][0],object[i].matrix[0][1], object[i].matrix[0][2], object[i].matrix[0][3],
							  object[i].matrix[1][0],object[i].matrix[1][1], object[i].matrix[1][2], object[i].matrix[1][3],
							  object[i].matrix[2][0],object[i].matrix[2][1], object[i].matrix[2][2], object[i].matrix[2][3],
							  object[i].matrix[3][0],object[i].matrix[3][1], object[i].matrix[3][2], object[i].matrix[3][3]);
		glm::mat3 N		= glm::transpose(glm::inverse(glm::mat3(MV)));
		glm::mat4 MVP	= P*MV;
		#else
		matrix_4x4_type N2,tmp, MVP;
		matrix_3x3_type N;
		MatrIdentity_4x4(N2);
		Fast_Inverse(object[i].matrix, tmp);
		MatrCopy_3x3_trsp(N2,tmp);
		MatrCopy_4x4_3x3(N2, N);		 
		MatrMul_4x4_4x4( object[i].matrix, P, MVP);
		#endif

		if(i==0)
			shader.Use();

		#ifdef USE_GLM
		glUniformMatrix3fv(shader("N"), 1, GL_FALSE, glm::value_ptr(N));
		glUniformMatrix4fv(shader("MV"), 1, GL_FALSE, glm::value_ptr(MV));		
		glUniformMatrix4fv(shader("MVP"), 1, GL_FALSE, glm::value_ptr(MVP)); 
		#else
		glUniformMatrix3fv(shader("N"), 1, GL_FALSE, &N[0][0]);
		glUniformMatrix4fv(shader("MV"), 1, GL_FALSE, &object[i].matrix[0][0]);		
		glUniformMatrix4fv(shader("MVP"), 1, GL_FALSE, &MVP[0][0]); 
		#endif
		if (object[i].id_texture!=-1) 
		{
			glBindTexture(GL_TEXTURE_2D, object[i].id_texture); // We set the active texture 		 
		}
		glUniform1i(shader("has_texture"), object[i].id_texture!=-1);
		
		glBindVertexArray(vaoID[i]);
			glDrawElements(GL_TRIANGLES, object[i].polygons_qty*3, GL_UNSIGNED_SHORT, 0);			
		glBindVertexArray(0);
	}
	shader.UnUse();
	glBindTexture(GL_TEXTURE_2D, 0);
	glutSwapBuffers();
}



void OnResize(int w, int h)
{
	glViewport (0, 0, (GLsizei) w, (GLsizei) h);
	//setup the projection matrix
	#ifdef USE_GLM
	P = glm::perspective(45.0f, (GLfloat)w/h, 5.0f, 10000.f);
	#else
	Perspective(45.0f, (GLfloat)w/h, 5.0f, 10000.f, P);
	#endif
}
void OnShutdown() { 
	for(int i=0;i<TOTAL_OBJECTS;i++){
		if(object[i].id_texture!=-1)
			glDeleteTextures(1, &object[i].id_texture);
	}
	glDeleteBuffers(TOTAL_OBJECTS, vboVerticesID);
	glDeleteBuffers(TOTAL_OBJECTS, vboTexCoordID);
	glDeleteBuffers(TOTAL_OBJECTS, vboNormalsID);
	glDeleteBuffers(TOTAL_OBJECTS, vboIndicesID);
	glDeleteVertexArrays(TOTAL_OBJECTS, vaoID);
}
void OnKey(unsigned char key, int x, int y)
{        
    switch (key)
	{  
        /*case ' ':
            rotation_x_increment=0;
            rotation_y_increment=0;
            rotation_z_increment=0;
        break;*/
		case 'j': case 'J':
			ObjTranslate(&object[obj_control],0,0,-1);
        break;
		case 'm': case 'M':
			ObjTranslate(&object[obj_control],0,0,1);
        break;
        case 'k': case 'K':
			ObjRotate(&object[obj_control],0,20,0);
        break;
        case 'l': case 'L':
			ObjRotate(&object[obj_control],0,-20,0);
        break;
        case 'r': case 'R':
            if (filling==0)
            {
                glPolygonMode (GL_FRONT_AND_BACK, GL_FILL); // Polygon rasterization mode (polygon filled)
                filling=1;
            }   
            else 
            {
                glPolygonMode (GL_FRONT_AND_BACK, GL_LINE); // Polygon rasterization mode (polygon outlined)
                filling=0;
            }
        break;
		case 27:
			exit(0);
		break;
    }
}
void OnSpecialKey(int key, int x, int y)
{        
    switch (key)
    {
         case GLUT_KEY_UP:
			ObjRotate(&object[obj_control],-20,0,0);
            //rotation_x_increment = rotation_x_increment +0.005;
        break;
        case GLUT_KEY_DOWN:
			ObjRotate(&object[obj_control],20,0,0);
            //rotation_x_increment = rotation_x_increment -0.005;
        break;
        case GLUT_KEY_LEFT:
			ObjRotate(&object[obj_control],0,0,20);
            //rotation_y_increment = rotation_y_increment +0.005;
        break;
        case GLUT_KEY_RIGHT:
			ObjRotate(&object[obj_control],0,0,-20);
            //rotation_y_increment = rotation_y_increment -0.005;
        break;
        case GLUT_KEY_PAGE_UP:
			obj_control++;
			if (obj_control>obj_qty) obj_control=0;
		break;
        case GLUT_KEY_PAGE_DOWN:
			obj_control--;
			if (obj_control<0) obj_control=obj_qty-1;
		break;
    }
}
void OnIdle() {
	rotation_x = rotation_x + rotation_x_increment;
    rotation_y = rotation_y + rotation_y_increment;
    rotation_z = rotation_z + rotation_z_increment;

    if (rotation_x > 359) rotation_x = 0;
    if (rotation_y > 359) rotation_y = 0;
    if (rotation_z > 359) rotation_z = 0;
	
	glutPostRedisplay();
}

void main(int argc, char** argv) {
	atexit(OnShutdown);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGB);	
	glutInitContextVersion (3, 3);
	glutInitContextFlags (GLUT_CORE_PROFILE | GLUT_DEBUG);
	glutInitWindowSize(screen_width, screen_height);
	#ifdef USE_GLM
	glutCreateWindow("Tutorial 6 - OpenGL 3.3 [Using GLM matrix library]");
	#else
	glutCreateWindow("Tutorial 6 - OpenGL 3.3 [Using spacesimulator.net matrix library.]");
	#endif

	glewExperimental = GL_TRUE;
	GLenum err = glewInit();
	if (GLEW_OK != err)	{
		cerr<<"Error: "<<glewGetErrorString(err)<<endl;
	} else {
		if (GLEW_VERSION_3_3)
		{
			cout<<"Driver supports OpenGL 3.3\nDetails:"<<endl;
		}
	}
	cout<<"Using GLEW "<<glewGetString(GLEW_VERSION)<<endl;
	cout<<"Vendor: "<<glGetString (GL_VENDOR)<<endl;
	cout<<"Renderer: "<<glGetString (GL_RENDERER)<<endl;
	cout<<"Version: "<<glGetString (GL_VERSION)<<endl;
	cout<<"GLSL: "<<glGetString (GL_SHADING_LANGUAGE_VERSION)<<endl;
	InitGL();
	glutDisplayFunc(OnRender);
	glutReshapeFunc(OnResize);
	glutKeyboardFunc(OnKey);
	glutSpecialFunc(OnSpecialKey);
	glutIdleFunc(OnIdle);
	glutMainLoop();
}